﻿using OUDBCommonLibs.Interfaces;
using OUDBCommonLibs.Query;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using OUDBCommonLibs.ExtensionMethods;
using OUDBCommonLibs.IntervalTree;
using System.Linq.Expressions;
using OUDBCommonLibs.Generators;
using OUDBCommonLibs.Variables;
using OUDBCommonLibs.Predicates;
using OUDBCommonLibs.IntervalTree;

namespace OUDBCache.Cache.SemanticCache
{
    public class SimulatedSimpleSingleAttributeCache<RT,AT> : SemanticCacheBase<RT>, ISemanticQueryGenerator<RT>
        where RT : new()
        where AT : struct, IComparable<AT>
    {
        private Expression<Func<RT, AT>> _attributeExpression;

        public SimulatedSimpleSingleAttributeCache(Expression<Func<RT, AT>> attributeExpression, ICachedQueryBatchExperiment<RT> experiment)
            :base(experiment.MaxCachedTuples, experiment.MaxCacheRegions)
        {
            this.Experiment = experiment;
            this._attributeExpression = attributeExpression;
        }

        public ICachedQueryBatchExperiment<RT> Experiment { get; set; }
        public override ISemanticCacheTrimmedQuery<RT, PT> TrimQuery<PT>(IQuery<RT, PT> query)
        {
            throw new NotImplementedException();
        }

        public override void AddToCache<PT>(ISemanticCacheRegion<RT, PT> region)
        {
            region.LastUsed = DateTime.Now;
            this.SemanticRegions.Add(region);

            while (this.SemanticRegions.Count > this.Experiment.MaxCacheableRegions && this.SemanticRegions.Count > 0) this.SemanticRegions.Remove(this.SemanticRegions.OrderByDescending(r => r.LastUsed).First());
        }

        public IQuery<RT> GetCacheHit(RandomNumericVariableBase percentOfRelationVariable)
        {
            double percentOfRelation = percentOfRelationVariable.Sample();

            ISemanticCacheRegion<RT> region = this.SemanticRegions.OrderBy(x => Math.Abs(percentOfRelation - x.UnTypedQuery.Predicate.PercentOfRelation())).First();

            region.LastUsed = DateTime.Now;

            return region.UnTypedQuery;
        }

        public IQuery<RT> GetCacheMiss(RandomNumericVariableBase percentOfRelationVariable)
        {
            double percentOfRelation = percentOfRelationVariable.Sample();

            IList<Interval<double>> intervals = new List<Interval<double>>();

            foreach (ISemanticCacheRegion<RT> region in this.SemanticRegions)
            {
                IList<Interval<double>> regionIntervals = region.UnTypedQuery.Predicate.ComputeIntervals<RT,double>();

                intervals = intervals.CombineIntervals(regionIntervals);
            }

            IQuantile<double> quantile = this._attributeExpression.ResolveAttributeDesc().ResolveQuantileForSingleAttributeNumeric();
            ComparableAttributePredicateGenerator<RT> attrPredGenerator = new ComparableAttributePredicateGenerator<RT>();
            IPredicate<RT> pred;

            if (intervals.Count > 0)
            {
                IList<Interval<double>> probabilityIntervals = new List<Interval<double>>();

                foreach (Interval<double> valueInterval in intervals)
                {
                    double startProb = quantile.CalculateProbability(valueInterval.Start);
                    double endProb = quantile.CalculateProbability(valueInterval.End);

                    probabilityIntervals.Add(new Interval<double>(startProb, endProb));
                }

                IList<Interval<double>> negatedProbabilityIntervals = probabilityIntervals.NegateIntervals(0.0, 1.0);

                List<Interval<double>> missProbabilityIntervals = negatedProbabilityIntervals.Where(i => i.End - i.Start > percentOfRelation).ToList();

                Interval<double> missProbabilityInterval;

                if (missProbabilityIntervals.Count > 1)
                {
                    UniformVariable intervalSelector = new UniformVariable(0, missProbabilityIntervals.Count);
                    int ndx = (int)intervalSelector.Sample();

                    missProbabilityInterval = missProbabilityIntervals[ndx];
                }
                else
                {
                    missProbabilityInterval = missProbabilityIntervals.First();
                }

                UniformVariable startPosSelector = new UniformVariable(missProbabilityInterval.Start, missProbabilityInterval.End - percentOfRelation);
                pred = attrPredGenerator.GenerateAttributeConjunctionRangePredicate<AT>(this._attributeExpression, new ConstantVariable(percentOfRelation), startPosSelector);
            }
            else
            {
                UniformVariable startPosSelector = new UniformVariable(0.0, 1.0 - percentOfRelation);
                pred = attrPredGenerator.GenerateAttributeConjunctionRangePredicate<AT>(this._attributeExpression, new ConstantVariable(percentOfRelation), startPosSelector);
            }

            IQuery<RT, RT> query = new SimpleQuery<RT>(pred);

            this.AddToCache(new SemanticCacheRegion<RT, RT>(query, null));

            return query; 
        }

        public IQuery<RT> GetCachePartialHit(RandomNumericVariableBase percentOfRelationVariable, RandomNumericVariableBase percentOfCacheOverlapVariable)
        {
            double percentOfRelation = percentOfRelationVariable.Sample();
            double percentOfCacheOverlap = percentOfCacheOverlapVariable.Sample();

            IList<Interval<double>> intervals = new List<Interval<double>>();

            foreach (ISemanticCacheRegion<RT> region in this.SemanticRegions)
            {
                IList<Interval<double>> regionIntervals = region.UnTypedQuery.Predicate.ComputeIntervals<RT, double>();

                intervals = intervals.CombineIntervals(regionIntervals);
            }

            IQuantile<double> quantile = this._attributeExpression.ResolveAttributeDesc().ResolveQuantileForSingleAttributeNumeric();
            ComparableAttributePredicateGenerator<RT> attrPredGenerator = new ComparableAttributePredicateGenerator<RT>();

            IList<Interval<double>> probabilityIntervals = new List<Interval<double>>();

            foreach (Interval<double> valueInterval in intervals)
            {
                double startProb = quantile.CalculateProbability(valueInterval.Start);
                double endProb = quantile.CalculateProbability(valueInterval.End);

                probabilityIntervals.Add(new Interval<double>(startProb, endProb));
            }

            IList<Interval<double>> negatedProbabilityIntervals = probabilityIntervals.NegateIntervals(0.0, 1.0);

            double cachedPercentOfRelation = percentOfRelation * percentOfCacheOverlap;
            double unCachedPercentOfRelation = percentOfRelation * (1.0 - percentOfCacheOverlap);
            List<CachedProbabilityInterval> cachedIntervals = probabilityIntervals.ToCachedProbabilityIntervals().Where(cp => cp.Size > cachedPercentOfRelation).ToList();
            List<UnCachedProbabilityInterval> unCachedIntervals = negatedProbabilityIntervals.ToUnCachedProbabilityIntervals().Where(ucp => ucp.Size > unCachedPercentOfRelation).ToList();

            CachedProbabilityInterval cacheInterval;
            UnCachedProbabilityInterval unCacheInterval;

            if (cachedIntervals.Count > 1)
            {
                UniformVariable cacheSelector = new UniformVariable(0, cachedIntervals.Count);
                int ndx = (int)cacheSelector.Sample();
                cacheInterval = cachedIntervals[ndx];
            }
            else
            {
                cacheInterval = cachedIntervals.First();
            }

            if (unCachedIntervals.Count > 1)
            {
                UniformVariable unCacheSelector = new UniformVariable(0, unCachedIntervals.Count);
                int ndx = (int)unCacheSelector.Sample();
                unCacheInterval = unCachedIntervals[ndx];
            }
            else
            {
                unCacheInterval = unCachedIntervals.First();
            }

            UniformVariable cacheStartPosSelector = new UniformVariable(cacheInterval.Interval.Start, cacheInterval.Interval.End - cachedPercentOfRelation);
            UniformVariable unCacheStartPosSelector = new UniformVariable(unCacheInterval.Interval.Start, unCacheInterval.Interval.End - unCachedPercentOfRelation);

            IPredicate<RT> cachedPred = attrPredGenerator.GenerateAttributeConjunctionRangePredicate<AT>(this._attributeExpression, new ConstantVariable(cachedPercentOfRelation), cacheStartPosSelector);
            IPredicate<RT> unCachedPred = attrPredGenerator.GenerateAttributeConjunctionRangePredicate<AT>(this._attributeExpression, new ConstantVariable(unCachedPercentOfRelation), unCacheStartPosSelector);

            IBinaryPredicate<RT> pred = new BinaryPredicate<RT>(cachedPred, unCachedPred, BinaryComparisonType.OR);

            IQuery<RT, RT> remainderQuery = new SimpleQuery<RT>(unCachedPred);
            this.AddToCache(new SemanticCacheRegion<RT, RT>(remainderQuery, null));

            return new SimpleQuery<RT>(pred);
        }

        public IQuery<RT> GetCacheExtendedHit(RandomNumericVariableBase percentOfRelationVariable)
        {
            IList<Interval<double>> intervals = new List<Interval<double>>();

            foreach (ISemanticCacheRegion<RT> region in this.SemanticRegions)
            {
                IList<Interval<double>> regionIntervals = region.UnTypedQuery.Predicate.ComputeIntervals<RT, double>();

                intervals = intervals.CombineIntervals(regionIntervals);
            }

            IQuantile<double> quantile = this._attributeExpression.ResolveAttributeDesc().ResolveQuantileForSingleAttributeNumeric();
            ComparableAttributePredicateGenerator<RT> attrPredGenerator = new ComparableAttributePredicateGenerator<RT>();

            IList<Interval<double>> probabilityIntervals = new List<Interval<double>>();

            foreach (Interval<double> valueInterval in intervals)
            {
                double startProb = quantile.CalculateProbability(valueInterval.Start);
                double endProb = quantile.CalculateProbability(valueInterval.End);

                probabilityIntervals.Add(new Interval<double>(startProb, endProb));
            }

            double percentOfRelation;
            List<CachedProbabilityInterval> cachedIntervals;
            do
            {
                percentOfRelation = percentOfRelationVariable.Sample();
                cachedIntervals = probabilityIntervals.ToCachedProbabilityIntervals().Where(cp => cp.Size > percentOfRelation).ToList();
            } while (cachedIntervals.Count == 0);
           
            CachedProbabilityInterval cacheInterval;

            
            if (cachedIntervals.Count > 1)
            {
                UniformVariable cacheSelector = new UniformVariable(0, cachedIntervals.Count);
                int ndx = (int)cacheSelector.Sample();
                cacheInterval = cachedIntervals[ndx];
            }
            else
            {
                cacheInterval = cachedIntervals.First();
            }

            UniformVariable cacheStartPosSelector = new UniformVariable(cacheInterval.Interval.Start, cacheInterval.Interval.End - percentOfRelation);
           
            IPredicate<RT> pred = attrPredGenerator.GenerateAttributeConjunctionRangePredicate<AT>(this._attributeExpression, new ConstantVariable(percentOfRelation), cacheStartPosSelector);

            IList<Interval<double>> predIntervals = pred.ComputeIntervals<RT, double>();

            foreach (ISemanticCacheRegion<RT> region in this.SemanticRegions)
            {
                IList<Interval<double>> regionIntervals = region.UnTypedQuery.Predicate.ComputeIntervals<RT, double>();

                bool found = false;
                foreach (Interval<double> predInterval in predIntervals)
                {
                    foreach (Interval<double> regionInterval in regionIntervals.Where(i => i.IsOverlaped(predInterval)))
                    {
                        Interval<double> intersectInterval = regionInterval.ComputeIntervalsIntersect(predInterval);

                        if (intersectInterval.Start == predInterval.Start && intersectInterval.End == predInterval.End)
                        {
                            found = true;
                            break;
                        }
                    }

                    if (found) break;

                }

                if (found) region.LastUsed = DateTime.Now;
            }

            return new SimpleQuery<RT>(pred);
        }

    }

    internal class ProbabilityInterval
    {
        internal ProbabilityInterval(Interval<double> interval) { this.Interval = interval; }
        internal Interval<double> Interval { get; private set; }
        internal double Size { get { return this.Interval.End - this.Interval.Start; } }
    }
    internal class CachedProbabilityInterval : ProbabilityInterval
    {
        internal CachedProbabilityInterval(Interval<double> interval) : base(interval) { }
    }

    internal class UnCachedProbabilityInterval : ProbabilityInterval
    {
        internal UnCachedProbabilityInterval(Interval<double> interval) : base(interval) { }
    }

    internal static class CacheHelper
    {
        internal static List<CachedProbabilityInterval> ToCachedProbabilityIntervals(this IEnumerable<Interval<double>> intervals)
        {
            List<CachedProbabilityInterval> retIntervals = new List<CachedProbabilityInterval>(intervals.Select(i => new CachedProbabilityInterval(i)));
            return retIntervals;
        }

        internal static List<UnCachedProbabilityInterval> ToUnCachedProbabilityIntervals(this IEnumerable<Interval<double>> intervals)
        {
            List<UnCachedProbabilityInterval> retIntervals = new List<UnCachedProbabilityInterval>(intervals.Select(i => new UnCachedProbabilityInterval(i)));
            return retIntervals;
        }
    }
}
