﻿using OUDBCommonLibs.Attributes;
using OUDBCommonLibs.Interfaces;
using OUDBCommonLibs.IntervalTree;
using OUDBCommonLibs.Predicates;
using OUDBCommonLibs.Tuple;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace OUDBCommonLibs.ExtensionMethods
{
    public static class ExpressionExtensionMethods
    {
        public static string ResolveAttributeName<RT, AT>(this Expression<Func<RT, AT>> attributeExpression)
        {
            return attributeExpression.ResolveAttributeDesc().Name;
        }

        public static ITupleAttributeDesc ResolveAttributeDesc<RT, AT>(this Expression<Func<RT, AT>> attributeExpression)
        {
            if (attributeExpression.NodeType == ExpressionType.Lambda)
            {
                LambdaExpression lambdaExpression = attributeExpression as LambdaExpression;

                if (lambdaExpression.Body.NodeType == ExpressionType.MemberAccess)
                {
                    MemberExpression memberExpression = lambdaExpression.Body as MemberExpression;
                    PropertyInfo member = memberExpression.Member as PropertyInfo;

                    return new TupleAttributeDesc(memberExpression.Member.Name, member.ReflectedType, member.PropertyType);
                }
                else if (lambdaExpression.Body.NodeType == ExpressionType.Convert)
                {
                    UnaryExpression unaryExpression = lambdaExpression.Body as UnaryExpression;

                    if (unaryExpression.Operand.NodeType == ExpressionType.MemberAccess)
                    {
                        MemberExpression memberExpression = unaryExpression.Operand as MemberExpression;
                        PropertyInfo member = memberExpression.Member as PropertyInfo;

                        return new TupleAttributeDesc(memberExpression.Member.Name, member.ReflectedType, member.PropertyType);
                    }
                }
            }


            throw new InvalidOperationException("Was not able to resolve attribute name from given expression");
        }

        public static string ResolveRelationName(this Type type)
        {
            TupleAttribute attr = type.GetCustomAttribute(typeof(TupleAttribute), true) as TupleAttribute;

            return attr.RelationName;

            throw new InvalidOperationException(string.Format("Was not able to resolve relation name from given relation for type {0}", type.FullName));
        }

        public static string ResolveRelationName<RT>(this IRelation<RT> relation)
        {
            Type type = typeof(RT);

            return type.ResolveRelationName();
        }

        public static string TupleToString<T>(this T tuple)
        {
            List<string> propertiesStrings = new List<string>();
            Type type = typeof(T);
            foreach (PropertyInfo pInfo in type.GetProperties())
            {
                Attribute attr = pInfo.GetCustomAttribute(typeof(TuplePropertyAttribute));

                if (attr != null)
                {
                    object value = pInfo.GetGetMethod().Invoke(tuple, null);

                    propertiesStrings.Add(string.Format("{0} => {1}", pInfo.Name, value));
                }
            }

            return string.Join(", ", propertiesStrings.ToArray());
        }

        public static RandomAttributeSchemaBase ResolveAttributeSchema<RT, AT>(this Expression<Func<RT, AT>> attributeExpression)
        {
            ITupleAttributeDesc attrDesc = attributeExpression.ResolveAttributeDesc();

            return attrDesc.ResolveAttributeSchema();
        }
        public static RandomAttributeSchemaBase ResolveAttributeSchema(this ITupleAttributeDesc attrDesc)
        {
            RandomTuplePropertyAttribute tuplePropertyAttribute = attrDesc.RelationType.GetProperty(attrDesc.Name).GetCustomAttribute(typeof(RandomTuplePropertyAttribute)) as RandomTuplePropertyAttribute;

            if (tuplePropertyAttribute != null)
            {
                return tuplePropertyAttribute.AttributeSchema;
            }

            throw new InvalidOperationException(string.Format("Could not resolve attribute schema for given expression"));
        }

        public static IQuantile<double> ResolveQuantileForSingleAttributeNumeric(this ITupleAttributeDesc attrDesc)
        {
            RandomAttributeSchemaBase schemaBase = attrDesc.ResolveAttributeSchema();


            if (!(schemaBase is RandomNumbericAttributeSchema)) throw new InvalidOperationException("Non-numeric schemas not supported");

            RandomNumbericAttributeSchema numericSchema = schemaBase as RandomNumbericAttributeSchema;
            IQuantile<double> quantile = numericSchema.Variable as IQuantile<double>;

            return quantile;
        }

        public static IQuantile<double> ResolveQuantileForSingleAttributeNumeric<RT>(this IPredicate<RT> predicate)
        {
            IList<ITupleAttributeDesc> predAttrs = predicate.GetPredicateAttributes<RT>();

            if (predAttrs.Count == 0) throw new InvalidOperationException("No attributes found in predicate");
         
            ITupleAttributeDesc attrDesc = predAttrs.First();

            foreach (ITupleAttributeDesc desc in predAttrs)
            {
                if (attrDesc.RelationType != desc.RelationType ||
                    attrDesc.Name != desc.Name ||
                    attrDesc.AttributeType != desc.AttributeType)
                {
                    throw new InvalidOperationException("Multiple Attributes not supported");
                }
            }

            return attrDesc.ResolveQuantileForSingleAttributeNumeric();
        }

        public static double PercentOfRelation<RT>(this IPredicate<RT> predicate)
        {
            IQuantile<double> quantile = predicate.ResolveQuantileForSingleAttributeNumeric();

            IList<Interval<double>> intervals = predicate.ComputeIntervals<RT, double>();

            double percentOfRelation = 0.0;
            foreach (Interval<double> interval in intervals)
            {
                double start = quantile.CalculateProbability(interval.Start);
                double end = quantile.CalculateProbability(interval.End);

                percentOfRelation += end - start;
            }

            return percentOfRelation;
        }

        public static bool IsNumericType(this Type type)
        {
            if (type == typeof(double)) return true;
            if (type == typeof(float)) return true;
            if (type == typeof(int)) return true;

            return false;
        }
    }
}
