﻿using OUDBCommonLibs.Interfaces;
using OUDBCommonLibs.IntervalTree;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using OUDBCommonLibs.ExtensionMethods;
using OUDBCommonLibs.Predicates;
using OUDBCommonLibs.Attributes;
using System.Linq.Expressions;
using OUDBCommonLibs.IntervalTree;

namespace OUDBCommonLibs.ExtensionMethods
{
    public static class IntervalExtensionMethods
    {
        public static bool IsInInterval<RT,AT>(this IPredicate<RT> predicate, AT value)
            where AT : struct, IComparable<AT>
        {
            if (predicate is IBinaryPredicate<RT>)
            {
                IBinaryPredicate<RT> boolPred = predicate as IBinaryPredicate<RT>;

                bool left = boolPred.Left.IsInInterval(value);
                bool right = boolPred.Right.IsInInterval(value);

                switch (boolPred.Type)
                {
                    case BinaryComparisonType.AND:
                        return left && right;


                    case BinaryComparisonType.OR:
                    default:
                        return left || right;
                }
            }
            else if (predicate is NegatePredicate<RT>)
            {
                NegatePredicate<RT> negPred = predicate as NegatePredicate<RT>;

                bool pred = negPred.Predicate.IsInInterval(value);

                return !pred;
            }
            else if (predicate is IComparableAttributePredicate<RT, AT>)
            {
                IComparableAttributePredicate<RT, AT> compPred = predicate as IComparableAttributePredicate<RT, AT>;

                int compare = compPred.CompareValue.CompareTo(value);

                switch (compPred.Type)
                {
                    case ComparisonType.GREATER_THAN:
                        return compare > 0;

                    case ComparisonType.GREATER_THAN_OR_EQUAL:
                        return compare >= 0;

                    case ComparisonType.LESS_THAN:
                        return compare < 0;

                    case ComparisonType.LESS_THAN_OR_EQUAL:
                        return compare <= 0;

                    case ComparisonType.EQUAL:
                    default:
                        return compare == 0;
                }
            }
            else
            {
                return false;
            }
        }

        public static IList<Interval<AT>> ComputeIntervals<RT,AT>(this IPredicate<RT> predicate)
            where AT : struct, IComparable<AT>
        {
            List<Interval<AT>> intervals = new List<Interval<AT>>();

            RandomNumbericAttributeSchema randomNumericSchema = predicate.GetPredicateAttributes<RT>().First().ResolveAttributeSchema() as RandomNumbericAttributeSchema;

            IQuantile<AT> quantile = randomNumericSchema.Variable as IQuantile<AT>;

            AT typedMin = quantile.CalculateValue(0);
            AT typedMax = quantile.CalculateValue(1);

            Action<ComparisonType, AT> mergeComparableIntervals = (ComparisonType type, AT value) =>
            {
                switch (type)
                {
                    case ComparisonType.GREATER_THAN:
                    case ComparisonType.GREATER_THAN_OR_EQUAL:
                        intervals.Add(new Interval<AT>(value, typedMax));
                        break;

                    case ComparisonType.LESS_THAN:
                    case ComparisonType.LESS_THAN_OR_EQUAL:
                        intervals.Add(new Interval<AT>(typedMin, value));
                        break;

                    case ComparisonType.EQUAL:
                    default:
                        intervals.Add(new Interval<AT>(value, value));
                        break;
                }
            };

            if (predicate is IBinaryPredicate<RT>)
            {
                IBinaryPredicate<RT> boolPred = predicate as IBinaryPredicate<RT>;

                IList<Interval<AT>> leftIntervals = boolPred.Left.ComputeIntervals<RT, AT>();
                IList<Interval<AT>> rightIntervals = boolPred.Right.ComputeIntervals<RT, AT>();

                
                switch (boolPred.Type)
                {
                    case BinaryComparisonType.AND:
                        IntervalTree<AT> rightIntervalTree = new IntervalTree<AT>();
                        foreach (Interval<AT> interval in rightIntervals) rightIntervalTree.Add(interval);

                        List<Interval<AT>> allIntersections = new List<Interval<AT>>();
                        Dictionary<Interval<AT>, List<Interval<AT>>> leftIntersectingIntervals = new Dictionary<Interval<AT>,List<Interval<AT>>>();
                        foreach (Interval<AT> left in leftIntervals)
                        {
                            List<Interval<AT>> rightIntersectingIntervals = rightIntervalTree.Search(left);

                            if (rightIntersectingIntervals.Count != 0)
                            {
                                List<Interval<AT>> intersections = new List<Interval<AT>>();
                                foreach (Interval<AT> right in rightIntersectingIntervals)
                                {
                                    Interval<AT> tmp = left.ComputeIntervalsIntersect(right);
                                    intervals.Add(tmp);
                                }
                            }
                        }
                        break;

                    case BinaryComparisonType.OR:
                    default:
                        intervals.AddRange(leftIntervals.CombineIntervals(rightIntervals));
                        break;
                }
            }
            else if (predicate is NegatePredicate<RT>)
            {
                NegatePredicate<RT> negPred = predicate as NegatePredicate<RT>;

                IList<Interval<AT>> negateIntervals = negPred.Predicate.ComputeIntervals<RT,AT>();

                AT currMin = typedMin;
                foreach (Interval<AT> negate in negateIntervals)
                {
                    if (negate.Start.CompareTo(currMin) > 0)
                    {
                        //found one, add interval before upto this
                        intervals.Add(new Interval<AT>(currMin, negate.Start));
                        currMin = negate.End;
                    }
                }

                intervals.Add(new Interval<AT>(currMin, typedMax));

            }
            else if (predicate is IComparableAttributePredicate<RT, AT>)
            {
                IComparableAttributePredicate<RT, AT> compPred = predicate as IComparableAttributePredicate<RT, AT>;

                mergeComparableIntervals(compPred.Type, (AT)Convert.ChangeType(compPred.CompareValue, typeof(AT)));
            }
            else if (predicate is IComparableAttributePredicate<RT, int> && typeof(AT).IsNumericType())
            {
                IComparableAttributePredicate<RT, int> compPred = predicate as IComparableAttributePredicate<RT, int>;

                mergeComparableIntervals(compPred.Type, (AT)Convert.ChangeType(compPred.CompareValue, typeof(AT)));
            }
            else if (predicate is IComparableAttributePredicate<RT, float> && typeof(AT).IsNumericType())
            {
                IComparableAttributePredicate<RT, float> compPred = predicate as IComparableAttributePredicate<RT, float>;

                mergeComparableIntervals(compPred.Type, (AT)Convert.ChangeType(compPred.CompareValue, typeof(AT)));
            }
            else if (predicate is IComparableAttributePredicate<RT, double> && typeof(AT).IsNumericType())
            {
                IComparableAttributePredicate<RT, double> compPred = predicate as IComparableAttributePredicate<RT, double>;

                mergeComparableIntervals(compPred.Type, (AT)Convert.ChangeType(compPred.CompareValue, typeof(AT)));
            }

            return intervals.CollapseIntervals();
        }

        public static IList<Interval<AT>> CombineIntervals<AT>(this IList<Interval<AT>> left, IList<Interval<AT>> right)
           where AT : struct, IComparable<AT>
        {
            foreach (Interval<AT> interval in right)
            {
                left.Add(interval);
            }

            return left.CollapseIntervals();
        }

        public static IList<Interval<AT>> NegateIntervals<AT>(this IList<Interval<AT>> intervals, AT min, AT max)
            where AT : struct, IComparable<AT>
        {
            IList<Interval<AT>> negatedIntervals = new List<Interval<AT>>();
            AT intervalMin = min;
            AT intervalMax = max;

            AT currMin = intervalMin;
            foreach (Interval<AT> negate in intervals)
            {
                if (negate.Start.CompareTo(currMin) > 0)
                {
                    //found one, add interval before upto this
                    negatedIntervals.Add(new Interval<AT>(currMin, negate.Start));
                }

                currMin = negate.End;
            }


            negatedIntervals.Add(new Interval<AT>(currMin, intervalMax));

            return negatedIntervals;
        }

        public static IList<Interval<AT>> CollapseIntervals<AT>(this IList<Interval<AT>> intervals)
            where AT: struct, IComparable<AT>
        {
            Stack<Interval<AT>> stack = new Stack<Interval<AT>>();

            if (intervals.Count <= 0) return stack.ToList();

            IList<Interval<AT>> sorted = intervals.SortIntervals();

            stack.Push(sorted[0]);

            for (int i = 1; i < sorted.Count; i++)
            {
                // get interval from stack top
                Interval<AT> top = stack.Peek();

                // if current interval is not overlapping with stack top,
                // push it to the stack
                if (top.End.CompareTo(sorted[i].Start) < 0)
                {
                    stack.Push(sorted[i]);
                }
                // Otherwise update the ending time of top if ending of current 
                // interval is more
                else if (top.End.CompareTo(sorted[i].End) < 0)
                {
                    top.End = sorted[i].End;
                    stack.Pop();
                    stack.Push(top);
                }
            }

            return stack.ToList();
        }

        public static IList<Interval<AT>> SortIntervals<AT>(this IList<Interval<AT>> intervals)
            where AT : struct, IComparable<AT>
        {
            Dictionary<AT, Interval<AT>> dict = new Dictionary<AT, Interval<AT>>();
            foreach (Interval<AT> interval in intervals)
            {
                dict.Add(interval.Start, interval);
            }

            List<AT> sorted = dict.Keys.ToList();
            sorted.Sort();

            IList<Interval<AT>> sortedRanges = new List<Interval<AT>>();

            foreach (AT key in sorted)
            {
                sortedRanges.Add(dict[key]);
            }

            return sortedRanges;
        }

        public static bool IsOverlaped<AT>(this Interval<AT> left, Interval<AT> right)
            where AT : struct, IComparable<AT>
        {
            bool compareA = left.Start.CompareTo(right.End) <= 0;
            bool compareB = right.Start.CompareTo(left.End) <= 0;

            return compareA && compareB;
        }

        public static Interval<AT> ComputeIntervalsIntersect<AT>(this Interval<AT> left, Interval<AT> right)
            where AT : struct, IComparable<AT>
        {
            if (!left.IsOverlaped(right)) throw new InvalidOperationException("Given intervals do not overlap");

            int compareStart = left.Start.CompareTo(right.Start);
            int compareEnd = left.End.CompareTo(right.End);

            AT intersectionStart = (compareStart < 0) ? right.Start : left.Start;
            AT intersectionEnd = (compareEnd < 0) ? left.End : right.End;

            return new Interval<AT>(intersectionStart, intersectionEnd);
        }

        public static bool OverlapsAny<AT>(this Interval<AT> interval, IEnumerable<Interval<AT>> intervals)
            where AT: struct, IComparable<AT>
        {
            return intervals.Any(i => IsOverlaped(interval, i));
        }

        public static IPredicate<RT> ToPredicate<RT, AT>(this IList<Interval<AT>> intervals, Expression<Func<RT, AT>> attributeExpression)
            where AT: struct, IComparable<AT>
        {
            IPredicate<RT> retPred = null;

            foreach (Interval<AT> interval in intervals)
            {
                IComparableAttributePredicate<RT, AT> attrLeftPred = new ComparableAttributePredicate<RT, AT>(attributeExpression, ComparisonType.GREATER_THAN, interval.Start);
                IComparableAttributePredicate<RT, AT> attrRightPred = new ComparableAttributePredicate<RT, AT>(attributeExpression, ComparisonType.LESS_THAN, interval.End);
                IBinaryPredicate<RT> pred = new BinaryPredicate<RT>(attrLeftPred, attrRightPred, BinaryComparisonType.AND);

                if (retPred == null)
                {
                    retPred = pred;
                }
                else
                {
                    retPred = new BinaryPredicate<RT>(retPred, pred, BinaryComparisonType.OR);
                }
            }

            return retPred;
        }

        public enum IntervalOverlap
        {
            BEFORE,
            WITHIN,
            AFTER
        }
    }
}
