﻿using OUDBCommonLibs.Attributes;
using OUDBCommonLibs.Interfaces;
using OUDBCommonLibs.Predicates;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;
using OUDBCommonLibs.ExtensionMethods;

namespace OUDBCommonLibs.Generators
{
    public class ComparableAttributePredicateGenerator<RT> : IComparableAttributeQueryPredicateGenerator<RT>
        where RT : new()
    {
        public ComparableAttributePredicateGenerator()
        {

        }

        public IPredicate<RT> GenerateAttributePredicate<AT>(Expression<Func<RT, AT>> attributeExpression, IVariable<double> percentVariable, bool useLowerPortion = true)
            where AT : IComparable<AT>
        {
            double percent = percentVariable.Sample();

            if (percent > 1.0 || percent < 0.0) throw new InvalidOperationException(string.Format("Invalid Percent({0}) for a GenerateAttributeQuery", percent));

            RandomNumbericAttributeSchema rndNumericAttr = attributeExpression.ResolveAttributeSchema() as RandomNumbericAttributeSchema;

            if (rndNumericAttr == null) throw new InvalidOperationException(string.Format("Generate query attribute must be RandomNumericalGeneratedAttributeDesc to generate queries on the attribute"));

            IQuantile<double> quantile = rndNumericAttr.Variable as IQuantile<double>;

            if (quantile == null) throw new InvalidOperationException(string.Format("RandomNumericalGeneratedAttributeDesc Variable must be IQuantile to generate queries on the attribute"));

            double quantilePercent = useLowerPortion ? percent : 1.0 - percent;

            return new ComparableAttributePredicate<RT, AT>(attributeExpression, useLowerPortion ? ComparisonType.LESS_THAN : ComparisonType.GREATER_THAN, (AT)Convert.ChangeType(quantile.CalculateValue(quantilePercent), typeof(AT)));
        }

        public IPredicate<RT> GenerateAttributeConjunctionRangePredicate<AT>(Expression<Func<RT, AT>> attributeExpression, IVariable<double> percentVariable, IVariable<double> startPercentVariable)
            where AT : IComparable<AT>
        {
            double percent = percentVariable.Sample();
            double startPercent = startPercentVariable.Sample();

            if (percent > 1.0 || percent < 0.0) throw new InvalidOperationException(string.Format("Invalid Percent({0}) for a GenerateAttributeConjunctionRangeQuery", percent));

            if ((startPercent + percent) > 1.0 || startPercent < 0.0) throw new InvalidOperationException(string.Format("Invalid Start Percent({0}) for a GenerateAttributeConjunctionRangeQuery", startPercent));

            RandomNumbericAttributeSchema rndNumericAttr = attributeExpression.ResolveAttributeSchema() as RandomNumbericAttributeSchema;

            if (rndNumericAttr == null) throw new InvalidOperationException(string.Format("Generate query attribute must be RandomNumericalGeneratedAttributeDesc to generate queries on the attribute"));

            IQuantile<AT> quantile = rndNumericAttr.Variable as IQuantile<AT>;

            if (quantile == null) throw new InvalidOperationException(string.Format("RandomNumericalGeneratedAttributeDesc Variable must be IQuantile to generate queries on the attribute"));


            IPredicate<RT> left = new ComparableAttributePredicate<RT, AT>(attributeExpression, ComparisonType.GREATER_THAN_OR_EQUAL, quantile.CalculateValue(startPercent));
            IPredicate<RT> right = new ComparableAttributePredicate<RT, AT>(attributeExpression, ComparisonType.LESS_THAN, quantile.CalculateValue(startPercent + percent));

            return new BinaryPredicate<RT>(left, right, BinaryComparisonType.AND);
        }

        public IPredicate<RT> GenerateAttributeDisjunctionRangePredicate<AT>(Expression<Func<RT, AT>> attributeExpression, IVariable<double> lowPercentVariable, IVariable<double> highPercentVariable)
             where AT : IComparable<AT>
        {
            double lowPercent = lowPercentVariable.Sample();
            double highPercent = highPercentVariable.Sample();

            if (lowPercent > 1.0 || lowPercent < 0.0) throw new InvalidOperationException(string.Format("Invalid Low Percent({0}) for a GenerateAttributeDisjunctionRangeQuery", lowPercent));

            if (highPercent > 1.0 || highPercent < 0.0) throw new InvalidOperationException(string.Format("Invalid High Percent({0}) for a GenerateAttributeDisjunctionRangeQuery", highPercent));

            if ((lowPercent + highPercent) > 1.0) throw new InvalidOperationException(string.Format("Invalid Low ({0}) / High ({1}) Percent for a GenerateAttributeDisjunctionRangeQuery", lowPercent, highPercent));

            RandomNumbericAttributeSchema rndNumericAttr = attributeExpression.ResolveAttributeSchema() as RandomNumbericAttributeSchema;

            if (rndNumericAttr == null) throw new InvalidOperationException(string.Format("Generate query attribute must be RandomNumericalGeneratedAttributeDesc to generate queries on the attribute"));

            IQuantile<AT> quantile = rndNumericAttr.Variable as IQuantile<AT>;

            if (quantile == null) throw new InvalidOperationException(string.Format("RandomNumericalGeneratedAttributeDesc Variable must be IQuantile to generate queries on the attribute"));


            IPredicate<RT> low = new ComparableAttributePredicate<RT, AT>(attributeExpression, ComparisonType.LESS_THAN, quantile.CalculateValue(lowPercent));

            IPredicate<RT> high = new ComparableAttributePredicate<RT, AT>(attributeExpression, ComparisonType.GREATER_THAN_OR_EQUAL, quantile.CalculateValue(1.0 - highPercent));

            return new BinaryPredicate<RT>(low, high, BinaryComparisonType.OR);
        }

        public IPredicate<RT> GenerateAttributeConjunctionDisjunctionRangePredicate<AT>(Expression<Func<RT, AT>> attributeExpression, IVariable<double> lowPercentVariable, IVariable<double> lowStartPercentVariable, IVariable<double> highPercentVariable, IVariable<double> highStartPercentVariable)
             where AT : IComparable<AT>
        {
            double lowPercent = lowPercentVariable.Sample();
            double lowStartPercent = lowPercentVariable.Sample();
            double highPercent = highPercentVariable.Sample();
            double highStartPercent = highStartPercentVariable.Sample();

            if (lowPercent > 1.0 || lowPercent < 0.0) throw new InvalidOperationException(string.Format("Invalid Low Percent({0}) for a GenerateAttributeConjunctionDisjunctionRangeQuery", lowPercent));

            if ((lowStartPercent + lowPercent) > 1.0 || lowStartPercent < 0.0) throw new InvalidOperationException(string.Format("Invalid Low Start Percent({0}) for a GenerateAttributeConjunctionDisjunctionRangeQuery", lowStartPercent));

            if (highPercent > 1.0 || highPercent < 0.0) throw new InvalidOperationException(string.Format("Invalid High Percent({0}) for a GenerateAttributeConjunctionDisjunctionRangeQuery", highPercent));

            if ((highStartPercent + highPercent) > 1.0 || highStartPercent < 0.0) throw new InvalidOperationException(string.Format("Invalid High Start Percent({0}) for a GenerateAttributeConjunctionDisjunctionRangeQuery", highStartPercent));

            if ((lowPercent + highPercent) > 1.0 || (lowPercent + lowStartPercent) > highStartPercent) throw new InvalidOperationException(string.Format("Invalid  Low Percent ({0}), High Start ({1}) and Low Start ({2}) combination  for a GenerateAttributeConjunctionDisjunctionRangeQuery", lowPercent, highStartPercent, lowStartPercent));

            RandomNumbericAttributeSchema rndNumericAttr = attributeExpression.ResolveAttributeSchema() as RandomNumbericAttributeSchema;

            if (rndNumericAttr == null) throw new InvalidOperationException(string.Format("Generate query attribute must be RandomNumericalGeneratedAttributeDesc to generate queries on the attribute"));

            IQuantile<AT> quantile = rndNumericAttr.Variable as IQuantile<AT>;

            if (quantile == null) throw new InvalidOperationException(string.Format("RandomNumericalGeneratedAttributeDesc Variable must be IQuantile to generate queries on the attribute"));

            IPredicate<RT> lowLeft = new ComparableAttributePredicate<RT, AT>(attributeExpression, ComparisonType.GREATER_THAN_OR_EQUAL, quantile.CalculateValue(lowStartPercent));
            IPredicate<RT> lowRight = new ComparableAttributePredicate<RT, AT>(attributeExpression, ComparisonType.LESS_THAN, quantile.CalculateValue(lowStartPercent + lowPercent));
            IPredicate<RT> low = new BinaryPredicate<RT>(lowLeft, lowRight, BinaryComparisonType.AND);

            IPredicate<RT> highLeft = new ComparableAttributePredicate<RT, AT>(attributeExpression, ComparisonType.GREATER_THAN_OR_EQUAL, quantile.CalculateValue(highStartPercent));
            IPredicate<RT> highRight = new ComparableAttributePredicate<RT, AT>(attributeExpression, ComparisonType.LESS_THAN, quantile.CalculateValue(highStartPercent + highPercent));
            IPredicate<RT> high = new BinaryPredicate<RT>(highLeft, highRight, BinaryComparisonType.AND);

            return new BinaryPredicate<RT>(low, high, BinaryComparisonType.OR);
        }
    }
}
