﻿using OUDBCommonLibs.Interfaces;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;
using OUDBCommonLibs.ExtensionMethods;

namespace OUDBCommonLibs.Predicates
{
    public class ComparableAttributePredicate<RT, AT> : IComparableAttributePredicate<RT, AT>
        where AT : IComparable<AT>
    {
        public ComparableAttributePredicate(Expression<Func<RT, AT>> attributeExpression, ComparisonType type, AT compareValue)
        {
            this.AttributeExpression = attributeExpression;
            this.Type = type;
            this.CompareValue = compareValue;
        }

        private Expression<Func<RT, AT>> _attrExpression = null;
        private Func<RT, AT> _attrFunc = null;
        public Expression<Func<RT, AT>> AttributeExpression
        {
            get
            {
                return this._attrExpression;
            }

            private set
            {
                this._attrExpression = value;
                this._attrFunc = this._attrExpression.Compile();
            }
        }

        public ComparisonType Type
        {
            get;
            private set;
        }

        public AT CompareValue
        {
            get;
            private set;
        }

        public bool Match(RT tuple)
        {
            AT value = this._attrFunc(tuple);

            int compareResult = value.CompareTo(this.CompareValue);
            switch (this.Type)
            {
                case ComparisonType.GREATER_THAN:
                    return compareResult > 0;

                case ComparisonType.LESS_THAN:
                    return compareResult < 0;

                case ComparisonType.GREATER_THAN_OR_EQUAL:
                    return compareResult >= 0;

                case ComparisonType.LESS_THAN_OR_EQUAL:
                    return compareResult <= 0;

                case ComparisonType.EQUAL:
                default:
                    return compareResult == 0;
            }
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(this.AttributeExpression.ResolveAttributeName());

            switch (this.Type)
            {
                case ComparisonType.GREATER_THAN:
                    sb.Append(" > ");
                    break;

                case ComparisonType.LESS_THAN:
                    sb.Append(" < ");
                    break;

                case ComparisonType.GREATER_THAN_OR_EQUAL:
                    sb.Append(" >= ");
                    break;

                case ComparisonType.LESS_THAN_OR_EQUAL:
                    sb.Append(" <= ");
                    break;

                case ComparisonType.EQUAL:
                default:
                    sb.Append(" = ");
                    break;
            }

            sb.Append(this.CompareValue.ToString());

            return sb.ToString();
        }

        public bool Match(object tuple)
        {
            if (tuple is RT)
            {
                return this.Match((RT)tuple);
            }
            else
            {
                return false;
            }
        }

        public override bool Equals(object obj)
        {
            // If parameter is null return false.
            if (obj == null)
            {
                return false;
            }

            // If parameter cannot be cast return false.
            IComparableAttributePredicate<RT, AT> otherPred = obj as IComparableAttributePredicate<RT, AT>;
            if ((System.Object)otherPred == null)
            {
                return false;
            }

            // Return true if the fields match:
            return this.Equals(otherPred);
        }

        public bool Equals(IPredicate<RT> obj)
        {
            IComparableAttributePredicate<RT, AT> otherPred = obj as IComparableAttributePredicate<RT, AT>;

            // If parameter is null return false:
            if ((object)otherPred == null)
            {
                return false;
            }

            // Return true if the fields match:
            if (this.Type != otherPred.Type) return false;
            if (this.CompareValue.CompareTo(otherPred.CompareValue) != 0) return false;
            if (this.AttributeExpression.ResolveAttributeName() != otherPred.AttributeExpression.ResolveAttributeName()) return false;

            return true;
        }

        public override int GetHashCode()
        {
            var result = 0;
            result = (result * 397) ^ this.Type.GetHashCode();
            result = (result * 397) ^ this.CompareValue.GetHashCode();
            result = (result * 397) ^ this.AttributeExpression.ResolveAttributeName().GetHashCode();

            return result;
        }

        public IList<ITupleAttributeDesc> GetPredicateAttributes<RT>()
        {
            List<ITupleAttributeDesc> attributeDescs = new List<ITupleAttributeDesc>();

            attributeDescs.Add(this.AttributeExpression.ResolveAttributeDesc());

            return attributeDescs;
        }
    }
}
