﻿using OUDBCommonLibs.Generators;
using OUDBCommonLibs.Interfaces;
using OUDBCommonLibs.Variables;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;
using OUDBCommonLibs.ExtensionMethods;
using OUDBCommonLibs.Attributes;
using OUDBCommonLibs.Query;
using OUDBCache.Cache.SemanticCache;

namespace OUDBQueryGeneration.QueryBatchGenerator
{
    public class SingleAttributeCachedQueryBatchGenerator<RT, AT> : IQueryBatchGenerator<RT>
        where RT : new()
        where AT : struct, IComparable<AT>
    {
        private Expression<Func<RT, AT>> _attributeExpression;
        private SimulatedSimpleSingleAttributeCache<RT, AT> _cacheSimulator;
        public SingleAttributeCachedQueryBatchGenerator(ICachedQueryBatchExperiment<RT> experiment, Expression<Func<RT, AT>> attributeExpression, IEnumerable<ISemanticCacheRegion<RT,RT>> intitialCachedRegions = null)
        {
            this.Experiment = experiment;
            this._attributeExpression = attributeExpression;
            this._cacheSimulator = new SimulatedSimpleSingleAttributeCache<RT, AT>(this._attributeExpression, this.Experiment);

            if(intitialCachedRegions != null){
                foreach (ISemanticCacheRegion<RT,RT> cacheRegion in intitialCachedRegions)
                {
                    this._cacheSimulator.AddToCache(cacheRegion);
                }
            }
        }

        public ICachedQueryBatchExperiment<RT> Experiment { get; private set; }

        public IList<IQuery<RT>> GenerateQueryBatch(int size)
        {
            IList<IQuery<RT>> queryBatch;

            switch (this.Experiment.ExperimentalVariable)
            {
                case SemanticQueryType.HIT:
                    queryBatch = this.GenerateForCacheHitExperiment(size);
                    break;

                case SemanticQueryType.PARTIAL:
                    queryBatch = this.GeneratForPartialHitExperiment(size);
                    break;

                case SemanticQueryType.EXTENDED:
                    queryBatch = this.GeneratForExtendedHitExperiment(size);
                    break;


                default:
                    throw new InvalidOperationException("Unsupported Experimental Variable");

            }
            return queryBatch;
        }

        private IList<IQuery<RT>> GenerateForCacheHitExperiment(int size)
        {
            IList<IQuery<RT>> queryBatch = new List<IQuery<RT>>();
            int totalHit = 0;
            do
            {
                double currentHitPercent = (double)totalHit / (double)queryBatch.Count;

                if (currentHitPercent < this.Experiment.PercentExperimental && this._cacheSimulator.RegionsInCache > 0)
                {
                    //use a cached query for a hit
                    IQuery<RT> query = this._cacheSimulator.GetCacheHit(this.Experiment.QueryPercentOfTotalRelation);

                    queryBatch.Add(query);

                    Console.WriteLine(string.Format("Cache Hit:\n\tQuery:{0}", query.Predicate.ToString()));
                    totalHit++;
                }
                else
                {
                    double querySizePercent = this.Experiment.QueryPercentOfTotalRelation.Sample();

                    IQuery<RT> query = this._cacheSimulator.GetCacheMiss(this.Experiment.QueryPercentOfTotalRelation);

                    queryBatch.Add(query);

                    Console.WriteLine(string.Format("Cache Miss:\n\tQuery:{0}", query.Predicate.ToString()));
                }

            } while (queryBatch.Count < size);

            return queryBatch;
        }

        private IList<IQuery<RT>> GeneratForPartialHitExperiment(int size)
        {
            IList<IQuery<RT>> queryBatch = new List<IQuery<RT>>();
            int totalPartialHit = 0;
            do
            {
                double currentPartialHitPercent = (double)totalPartialHit / (double)queryBatch.Count;

                if (currentPartialHitPercent < this.Experiment.PercentExperimental && this._cacheSimulator.RegionsInCache > 0)
                {
                    //use a cached query for a hit
                    IQuery<RT> query = this._cacheSimulator.GetCachePartialHit(this.Experiment.QueryPercentOfTotalRelation, new ConstantVariable(this.Experiment.PercentSemanticHit));

                    queryBatch.Add(query);

                    Console.WriteLine(string.Format("Cache Hit:\n\tQuery:{0}", query.Predicate.ToString()));
                    totalPartialHit++;
                }
                else
                {
                    double querySizePercent = this.Experiment.QueryPercentOfTotalRelation.Sample();

                    IQuery<RT> query = this._cacheSimulator.GetCacheMiss(this.Experiment.QueryPercentOfTotalRelation);

                    queryBatch.Add(query);

                    Console.WriteLine(string.Format("Cache Miss:\n\tQuery:{0}", query.Predicate.ToString()));
                }

            } while (queryBatch.Count < size);

            return queryBatch;
        }

        private IList<IQuery<RT>> GeneratForExtendedHitExperiment(int size)
        {
            IList<IQuery<RT>> queryBatch = new List<IQuery<RT>>();
            int totalExtendedHit = 0;
            do
            {
                double currentPartialHitPercent = (double)totalExtendedHit / (double)queryBatch.Count;

                if (currentPartialHitPercent < this.Experiment.PercentExperimental && this._cacheSimulator.RegionsInCache > 0)
                {
                    //use a cached query for a hit
                    IQuery<RT> query = this._cacheSimulator.GetCacheExtendedHit(this.Experiment.QueryPercentOfTotalRelation);

                    queryBatch.Add(query);

                    Console.WriteLine(string.Format("Cache Hit:\n\tQuery:{0}", query.Predicate.ToString()));
                    totalExtendedHit++;
                }
                else
                {
                    double querySizePercent = this.Experiment.QueryPercentOfTotalRelation.Sample();

                    IQuery<RT> query = this._cacheSimulator.GetCacheMiss(this.Experiment.QueryPercentOfTotalRelation);

                    queryBatch.Add(query);

                    Console.WriteLine(string.Format("Cache Miss:\n\tQuery:{0}", query.Predicate.ToString()));
                }

            } while (queryBatch.Count < size);

            return queryBatch;
        }
    }
}
